import logging
from typing import Any

import numpy as np
from lion_core.session.branch import Branch
from lion_service import iModel
from lionfuncs import to_num
from pydantic import BaseModel, ConfigDict, Field, field_validator

from .config import DEFAULT_CHAT_CONFIG

PROMPT = "Please follow prompt and provide {num_scores} numeric score(s) in {score_range} for the given context. Return as {return_precision} format"


class ScoreModel(BaseModel):

    score: list | float = Field(
        default_factory=list,
        description="** A numeric score or a list of numeric scores.**",
    )

    model_config = ConfigDict(
        population_by_field_name=True,
        arbitrary_types_allowed=True,
    )

    @field_validator("score", mode="before")
    def validate_score(cls, value) -> list:
        return [value] if not isinstance(value, list) else value


async def score(
    score_range=(1, 10),
    instruction=None,
    guidance=None,
    context=None,
    system=None,
    reason: bool = False,
    actions: bool = False,
    tools: Any = None,
    imodel: iModel = None,
    branch: Branch = None,
    sender=None,
    recipient=None,
    clear_messages: bool = False,
    system_sender=None,
    system_datetime=None,
    return_branch=False,
    num_parse_retries: int = 0,
    retry_imodel: iModel = None,
    num_scores: int = 1,
    use_average: bool = False,
    precision: int = 0,
    default_score=np.nan,
    **kwargs,
) -> ScoreModel:

    if branch and branch.imodel:
        imodel = imodel or branch.imodel
    else:
        imodel = imodel or iModel(**DEFAULT_CHAT_CONFIG)

    branch = branch or Branch(imodel=imodel)

    return_precision = "integer" if precision == 0 else f"num:{precision}f"
    prompt = PROMPT.format(
        num_scores=num_scores,
        score_range=score_range,
        return_precision=return_precision,
    )
    if instruction:
        prompt = f"{instruction}\n\n{prompt} \n\n "

    if system:
        branch.add_message(
            system=system,
            system_datetime=system_datetime,
            sender=system_sender,
        )

    _context = [{"operation": prompt}]
    if context:
        _context.append(context)

    kwargs["frozen"] = False
    response = await branch.operate(
        instruction=instruction,
        guidance=guidance,
        context=_context,
        sender=sender,
        recipient=recipient,
        reason=reason,
        actions=actions,
        tools=tools,
        clear_messages=clear_messages,
        operative_model=ScoreModel,
        imodel=imodel,
        retry_imodel=retry_imodel,
        num_parse_retries=num_parse_retries,
        **kwargs,
    )

    return_kind = int if precision == 0 else float
    err = None
    try:
        if isinstance(response, dict):
            response = ScoreModel(**response)

        response.score = [
            to_num(
                i,
                upper_bound=score_range[1],
                lower_bound=score_range[0],
                num_type=return_kind,
                precision=precision,
                num_count=num_scores,
            )
            for i in response.score
        ]
        if use_average:
            scores = response.score
            scores = [scores] if not isinstance(scores, list) else scores
            response.score = np.mean(scores)

        if response.score and num_scores == 1:
            if isinstance(response.score, list):
                response.score = response.score[0]

        if return_branch:
            return response, branch
        return response

    except Exception as e:
        err = e
        pass

    logging.error(
        f"Error converting score to {return_kind}: {err}, "
        f"value is set to default: {default_score}"
    )
    response.score = default_score
    if return_branch:
        return response, branch
    return response



from typing import Any

from lion_core.operative.step_model import StepModel
from lion_core.session.branch import Branch
from lion_service import iModel
from pydantic import BaseModel, Field

from .config import DEFAULT_CHAT_CONFIG


class BrainstormModel(BaseModel):

    topic: str = Field(
        default_factory=str,
        description="**Specify the topic or theme for the brainstorming session.**",
    )
    ideas: list[StepModel] = Field(
        default_factory=list,
        description="**Provide a list of ideas needed to accomplish the objective. Each step should be as described in a `PlanStepModel`.**",
    )


PROMPT = "Please follow prompt and provide {num_steps} different ideas for the next step"


async def brainstorm(
    num_steps: int = 3,
    instruction=None,
    guidance=None,
    context=None,
    system=None,
    reason: bool = False,
    actions: bool = False,
    tools: Any = None,
    imodel: iModel = None,
    branch: Branch = None,
    sender=None,
    recipient=None,
    clear_messages: bool = False,
    system_sender=None,
    system_datetime=None,
    return_branch=False,
    num_parse_retries: int = 3,
    retry_imodel: iModel = None,
    branch_user=None,
    **kwargs,  # additional operate arguments
):
    if branch and branch.imodel:
        imodel = imodel or branch.imodel
    else:
        imodel = imodel or iModel(**DEFAULT_CHAT_CONFIG)

    prompt = PROMPT.format(num_steps=num_steps)

    branch = branch or Branch(imodel=imodel)
    if branch_user:
        branch.user = branch_user

    if system:
        branch.add_message(
            system=system,
            system_datetime=system_datetime,
            sender=system_sender,
        )
    _context = [{"operation": prompt}]
    if context:
        _context.append(context)

    response = await branch.operate(
        instruction=instruction,
        guidance=guidance,
        context=_context,
        sender=sender,
        recipient=recipient,
        reason=reason,
        actions=actions,
        tools=tools,
        clear_messages=clear_messages,
        operative_model=BrainstormModel,
        retry_imodel=retry_imodel,
        num_parse_retries=num_parse_retries,
        imodel=imodel,
        **kwargs,
    )
    if return_branch:
        return response, branch
    return response



from __future__ import annotations

import inspect
from collections.abc import Callable
from enum import Enum
from typing import Any

from lion_core.session.branch import Branch
from lion_service import iModel
from lionfuncs import string_similarity
from pydantic import BaseModel, Field

from lionagi.libs.sys_util import SysUtil

from .config import DEFAULT_CHAT_CONFIG

PROMPT = "Please select up to {max_num_selections} items from the following list {choices}. Provide the selection(s) into appropriate field in format required, and no comments from you"


class SelectionModel(BaseModel):
    selected: list[str | Enum] = Field(default_factory=list)


async def select(
    choices: list[str] | type[Enum],
    max_num_selections: int = 1,
    instruction=None,
    guidance=None,
    context=None,
    system=None,
    reason: bool = False,
    actions: bool = False,
    tools: Any = None,
    imodel: iModel = None,
    branch: Branch = None,
    sender=None,
    recipient=None,
    return_enum: bool = False,
    enum_parser: Callable = None,  # parse the model string response to appropriate type
    clear_messages: bool = False,
    system_sender=None,
    system_datetime=None,
    return_branch=False,
    num_parse_retries: int = 3,
    retry_imodel: iModel = None,
    branch_user=None,
    **kwargs,
) -> SelectionModel | tuple[SelectionModel, Branch]:

    if branch and branch.imodel:
        imodel = imodel or branch.imodel
    else:
        imodel = imodel or iModel(**DEFAULT_CHAT_CONFIG)

    selections = []
    if return_enum and not _is_enum(choices):
        raise ValueError("return_enum can only be True if choices is an Enum")

    if _is_enum(choices):
        selections = [selection.value for selection in choices]
    else:
        selections = choices

    prompt = PROMPT.format(
        max_num_selections=max_num_selections, choices=selections
    )

    if instruction:
        prompt = f"{instruction}\n\n{prompt} \n\n "

    branch = branch or Branch(imodel=imodel)
    if branch_user:
        try:
            a = SysUtil.get_id(branch_user)
            branch.user = a
        except:
            branch.user = branch_user
    if system:
        branch.add_message(
            system=system,
            system_datetime=system_datetime,
            system_sender=system_sender,
        )

    kwargs["frozen"] = False
    response_model: SelectionModel = await branch.operate(
        instruction=prompt,
        guidance=guidance,
        context=context,
        sender=sender,
        recipient=recipient,
        reason=reason,
        actions=actions,
        operative_model=SelectionModel,
        clear_messages=clear_messages,
        imodel=imodel,
        num_parse_retries=num_parse_retries,
        retry_imodel=retry_imodel,
        tools=tools,
        **kwargs,
    )

    selected = response_model
    if isinstance(response_model, BaseModel) and hasattr(
        response_model, "selected"
    ):
        selected = response_model.selected
    selected = [selected] if not isinstance(selected, list) else selected
    corrected_selections = [
        string_similarity(
            word=selection,
            correct_words=selections,
            return_most_similar=True,
        )
        for selection in selected
    ]

    if return_enum:
        out = []
        if not enum_parser:
            enum_parser = lambda x: x
        for selection in corrected_selections:
            selection = enum_parser(selection)
            for member in choices.__members__.values():
                if member.value == selection:
                    out.append(member)
        corrected_selections = out

    if isinstance(response_model, BaseModel):
        response_model.selected = corrected_selections

    elif isinstance(response_model, dict):
        response_model["selected"] = corrected_selections

    if return_branch:
        return response_model, branch
    return response_model


def _is_enum(choices):
    return inspect.isclass(choices) and issubclass(choices, Enum)



import asyncio
from typing import Any

import numpy as np
from lion_core.session.branch import Branch
from lion_core.session.session import Session
from lion_service import iModel
from lionfuncs import alcall, to_list

from .config import DEFAULT_CHAT_CONFIG
from .score import score

PROMPT = (
    "Given all items: \n {choices} \n\n Please follow prompt and give score "
    "to the item of interest: \n {item} \n\n"
)


async def rank(
    choices: list[Any],
    num_scorers: int = 5,
    instruction=None,
    guidance=None,
    context=None,
    system=None,
    reason: bool = False,
    actions: bool = False,
    tools: Any = None,
    imodel: iModel = None,
    branch: Branch = None,  # branch won't be used for the vote, it is for configuration
    clear_messages: bool = False,
    system_sender=None,
    system_datetime=None,
    num_parse_retries: int = 0,
    retry_imodel: iModel = None,
    return_session: bool = False,
    **kwargs,  # additional kwargs for score function
) -> dict:

    if branch and branch.imodel:
        imodel = imodel or branch.imodel
    else:
        imodel = imodel or iModel(**DEFAULT_CHAT_CONFIG)

    branch = branch or Branch(imodel=imodel)
    session = Session(default_branch=branch)

    async def _score(item):
        async with session.branches.async_lock:
            b_ = session.new_branch(messages=session.default_branch.messages)

        prompt = PROMPT.format(choices=choices, item=item)
        if instruction:
            prompt = f"{instruction}\n\n{prompt} \n\n "

        kwargs["branch"] = b_
        kwargs["score_range"] = kwargs.get("score_range", (1, 10))
        kwargs["num_scores"] = kwargs.get("num_scores", 1)
        kwargs["precision"] = kwargs.get("precision", 1)

        response = await score(
            instruction=prompt,
            guidance=guidance,
            context=context,
            system=system,
            system_datetime=system_datetime,
            system_sender=system_sender,
            sender=session.ln_id,
            recipient=b_.ln_id,
            default_score=-1,
            reason=reason,
            actions=actions,
            tools=tools,
            clear_messages=clear_messages,
            num_parse_retries=num_parse_retries,
            retry_imodel=retry_imodel,
            **kwargs,
        )

        if response.score == -1:
            return None

        return response

    async def _group_score(item):
        tasks = [asyncio.create_task(_score(item)) for _ in range(num_scorers)]
        responses = await asyncio.gather(*tasks)
        responses = [i for i in responses if i is not None]
        scores = to_list(
            [i.score for i in responses], dropna=True, flatten=True
        )
        return {
            "item": item,
            "scores": scores,
            "average": np.mean(scores) if scores else -1,
        }

    results = await alcall(choices, _group_score)
    results = sorted(results, key=lambda x: x["average"], reverse=True)
    if return_session:
        return results, session
    return results
